from pathlib import Path
from .conf import get_setting


def _get_lock_path(lockname: str) -> Path:
    cache_dir = get_setting("caerp.static_tmp", default="/tmp")
    directory = Path(cache_dir)
    return directory.joinpath(f"{lockname}.lock")


def is_locked(lockname):
    p = _get_lock_path(lockname)
    return p.exists()


def acquire_lock(lockname):
    p = _get_lock_path(lockname)
    p.touch()


def release_lock(lockname):
    p = _get_lock_path(lockname)
    p.unlink()



# -*- coding: utf-8 -*-
"""
Configuration management tools

Tools to retrieve elements from the main app configuration
"""
from caerp.models.config import get_admin_mail


def get_registry():
    """
    Return the Pyramid application registry associated to the current running
    application

    :returns: A Pyramid registry instance
    """
    from pyramid_celery import celery_app

    return celery_app.conf["PYRAMID_REGISTRY"]


def get_request():
    """
    Return the Pyramid application request associated to the current running
    application context
    """
    from pyramid_celery import celery_app

    return celery_app.conf["PYRAMID_REQUEST"]


def get_setting(key, mandatory=False, default=None):
    """
    Collect a given setting

    :param str key: The key to collect
    :param bool mandatory: Is the key mandatory (if it is a KeyError may be
    raised)
    :param default: The default value to return
    :rtype: str
    """
    settings = get_registry().settings
    if mandatory:
        return settings[key]
    else:
        return settings.get(key, default)


def get_sysadmin_mail():
    """
    Retrieve the sysadmin mail from the current configuration or None if not set

    :rtype: str
    """
    return get_setting("caerp.sysadmin_mail")


def get_recipients_addresses(request):
    """
    Retrieve recipients mail adresses
    :param obj request: The request object
    :returns: A list of mail adresses
    """
    _admin_mail = get_admin_mail()
    _sysadmin_mail = get_sysadmin_mail()
    mail_addresses = []
    if _admin_mail:
        mail_addresses.append(_admin_mail)
    if _sysadmin_mail:
        mail_addresses.append(_sysadmin_mail)

    if mail_addresses:
        setattr(request, "registry", get_registry())
    return mail_addresses



import locale

# flake8: noqa: E402

locale.setlocale(locale.LC_ALL, "fr_FR.UTF-8")
locale.setlocale(locale.LC_TIME, "fr_FR.UTF-8")

import logging
from pyramid.config import Configurator
from pyramid.paster import setup_logging
from pyramid.path import DottedNameResolver

from sqlalchemy import engine_from_config
from pyramid_beaker import set_cache_regions_from_settings

from caerp import (
    prepare_view_config,
    configure_filedepot,
)
from caerp.models import *
from caerp_base.models.initialize import initialize_sql
from caerp_celery.tasks.csv_import import (
    MODELS_CONFIGURATION as IMPORT_MODELS_CONFIGURATION,
)
from caerp_celery.tasks.export import MODELS_CONFIGURATION as EXPORT_MODELS_CONFIGURATION
from .interfaces import IAccountingOperationProducer, IAccountingFileParser


def configure_accounting_parser_service(config, settings):
    """
    Configure the service used to parse the accounting operations file (Grand livre)

    factory = request.find_service_factory(IAccountingFileParser)
    parser = factory(filepath, request)

    """
    for interface, default_module_path, setting_key in (
        (
            IAccountingFileParser,
            "caerp_celery.parsers.sage.parser_factory",
            "caerp_celery.interfaces.IAccountingFileParser",
        ),
        (
            IAccountingOperationProducer,
            "caerp_celery.parsers.sage.producer_factory",
            "caerp_celery.interfaces.IAccountingOperationProducer",
        ),
    ):
        if setting_key in settings:
            module_path = settings[setting_key]
        else:
            module_path = default_module_path

        service = DottedNameResolver().resolve(module_path)
        config.register_service_factory(service, interface)


def register_import_model(
    config, key, model, label, permission, excludes, callbacks=()
):
    """
    Register a model for import

    :param obj config: The pyramid configuration object
    :param str key: The key used to identify the model type
    :param class model: The model to be used
    :param str label: A label describing the type of datas
    :param str permission: The permission to associate to this import
    :param tuple excludes: The field of the model we don't want to handle in
    the import
    :param list callbacks: List of callables called after import action
    """
    IMPORT_MODELS_CONFIGURATION[key] = {
        "factory": model,
        "label": label,
        "permission": permission,
        "excludes": excludes,
        "callbacks": callbacks,
    }


def register_export_model(config, key, model, options={}):
    """
    Register a model for export

    :param obj config: The pyramid configuration object
    :param str key: The key used to identify the model type
    :param class model: The model to be used
    """
    EXPORT_MODELS_CONFIGURATION[key] = {"factory": model}
    EXPORT_MODELS_CONFIGURATION[key].update(options)


def includeme(config):
    """
    Includes some celery specific stuff in the main application
    """
    settings = config.get_settings()
    configure_accounting_parser_service(config, settings)
    config.add_directive("register_import_model", register_import_model)
    config.add_directive("register_export_model", register_export_model)
    config.include("caerp.views.invoices.routes")
    config.include("caerp_celery.tasks")


def worker(global_config, **settings):
    """
    Entry point for the pyramid celery stuff
    """
    logging.basicConfig()
    setup_logging(global_config["__file__"])
    logger = logging.getLogger(__name__)
    logger.info("Bootstraping app")
    engine = engine_from_config(settings, "sqlalchemy.")
    set_cache_regions_from_settings(settings)
    dbsession = initialize_sql(engine)

    from caerp import prepare_config
    # config.include('pyramid_celery')
    config = prepare_config(**settings)
    includeme(config)

    config.configure_celery(global_config["__file__"])
    config.commit()
    config = prepare_view_config(config, dbsession, from_tests=False, **settings)
    configure_filedepot(settings)
    config.include("caerp.views.files")
    return config.make_wsgi_app()


def scheduler(global_config, **settings):
    logging.basicConfig()
    setup_logging(global_config["__file__"])
    logger = logging.getLogger(__name__)
    logger.info("Bootstraping celery scheduler application")
    config = Configurator(settings=settings)
    config.include("pyramid_celery")
    config.configure_celery(global_config["__file__"])
    config.commit()
    return config.make_wsgi_app()



# -*- coding: utf-8 -*-
"""
Csv based importation module

1 - Load csv file get headers
1 - Load sql columns of the current model

2 - Try auto-association
3 - Generate form
4 - Populate form (with auto-associated columns and optionnal loaded
association)
5- insert or update elements
"""
import transaction
import csv

try:
    from io import StringIO
except ImportError:
    from io import StringIO
from collections import OrderedDict

from pyramid_celery import celery_app
from sqlalchemy import (
    Boolean,
    DateTime,
    Date,
    Integer,
    BigInteger,
    Float,
    Numeric,
)
from sqlalchemy.orm import (
    RelationshipProperty,
    exc as sqlalchemy_exc,
)
from sqlalchemy.exc import OperationalError
from sqlalchemy.schema import ColumnDefault
from sqla_inspect.base import BaseSqlaInspector

from caerp.compute import math_utils
from caerp.utils.datetimes import str_to_date
from caerp.utils import ascii
from caerp_celery.exception import (
    MissingMandatoryArgument,
    MultipleInstanceFound,
)
from caerp_celery.tasks import utils
from caerp_celery.models import (
    CsvImportJob,
)

logger = log = utils.get_logger(__name__)

MISSING_KEY_ERROR = "Erreur : Le champ {0} ({1}) est requis mais n'a pas été \
configuré à l'étape 2"
MULTIPLE_ENTRY_ERROR = "Pour une même valeur clé ({0}), plusieurs entrées \
ont été détectées, nous ne sommes pas en mesure de mettre à jour ces données."
NO_ENTRY_FOUND_ERROR = "Aucune entrée n'a pu être retrouvée pour la clé {0} \
avec la valeur {1}"
NO_ID_KEY_ERROR = "Une entrée n'a pas de valeur pour la clé {0}."
UNFILLED_VALUES = ("", None, 0)

BOOLEAN_FALSE = ("0", "false", "non", "", "False", "N", "Non", "NON", "NO")
DATETIME_FORMAT = "%d/%m/%Y"


DEFAULT_ID_LABEL = "Identifiant CAErp"
DEFAULT_DELIMITER = ";"
DEFAULT_QUOTECHAR = '"'
DELIMITERS = (";", ",", ":")
QUOTECHARS = ('"', '"')


MODELS_CONFIGURATION = {}


def format_input_value(value, sqla_column_dict, force_rel_creation=False):
    """
    format value to fetch the database storage expectations
    For example dates are input in dd/mm/yyyy format, we want datetime objects

    Relationship technics:

        ManyToOne relationship : we use the related_key (attribute of the
        related element) to find the related element we're supposed to point to

        OneToMany relationship: we use the related key (attribute of the
        related element) to instantiate a new related element. NB: It only
        works with related elements with one argument (typically list of dates
        ...)

    :param str value: The value coming from the csv file
    :param dict sqla_column_dict: The datas collected about the destination
    attribute
    :param bool force_rel_creation: Should we force the creation of a related
    element on import (default False), only in case of many to one
    relationships.  Note : the related_key attribute should be sufficient to
    create a new instance of the related element
    :returns: The formatted value
    :rtype: object (datetime) or string depending on the column
    """

    prop = sqla_column_dict["__col__"]

    res = value

    if isinstance(prop, RelationshipProperty):
        # Handle the relationship
        # Get the id of the corresponding model and return it
        if sqla_column_dict["rel_type"] == "manytoone":
            if "related_retriever" in sqla_column_dict:
                func = sqla_column_dict["related_retriever"]
                res = func(value, force_rel_creation)
            else:
                related_key = sqla_column_dict["related_key"]
                class_ = prop.mapper.class_
                # We query the database to get the corresponding element
                # filtering on the configured related_key
                res = (
                    class_.query().filter(getattr(class_, related_key) == value).first()
                )
                if res is None and force_rel_creation:
                    if value is not None and value.strip():
                        logger.debug(
                            "Creating a new element : %s %s" % (related_key, value)
                        )
                        creation_dict = {related_key: value}
                        res = class_(**creation_dict)
        else:
            # We have a one to many relationship, we generate an instance using
            # the related_key as instanciation attribute
            related_key = sqla_column_dict["related_key"]
            class_ = prop.mapper.class_
            if "formatter" in sqla_column_dict:
                value = sqla_column_dict["formatter"](value)
            args = {related_key: value}
            res = [class_(**args)]

    else:
        column = prop.columns[0]
        column_type = getattr(column.type, "impl", column.type)

        if "formatter" in sqla_column_dict:
            res = sqla_column_dict["formatter"](value)

        elif isinstance(column_type, Boolean):
            if value in BOOLEAN_FALSE:
                res = False
            else:
                res = True

        elif isinstance(
            column_type,
            (
                DateTime,
                Date,
            ),
        ):
            res = str_to_date(value)

        elif isinstance(column_type, (Integer, BigInteger)):
            res = math_utils.str_to_int(value, default=None)
        elif isinstance(column_type, (Numeric, Float)):
            res = math_utils.str_to_float(value, default=None)
    return res


class CsvImportAssociator(BaseSqlaInspector):
    """
    An importation tool used to associate csv datas to a destination dict

    Collect datas from a sqlalchemy model
    Generate a guessed association_dict from a list of field names

    for column in columns:
        if column:
            store_name_and_title
        else:
            store_name_title_and_values_if_appoption
    """

    config_key = "import"

    def __init__(self, model, excludes=()):
        BaseSqlaInspector.__init__(self, model)
        self.model = model
        self.excludes = excludes
        self.columns = self._collect_columns()

    def _collect_columns(self):
        """
        Collect the columns names and titles to be able to generate a
        importation form or guess associations
        """
        result = OrderedDict()
        todrop = []
        for prop in self.get_sorted_columns():
            if prop.key in self.excludes:
                continue
            info_dict = self.get_info_field(prop)
            colanderalchemy_infos = info_dict.get("colanderalchemy", {})

            import_dict = info_dict.get(self.config_key, {})

            ui_label = colanderalchemy_infos.get("title", prop.key)
            datas = {"name": prop.key, "label": ui_label, "__col__": prop}
            datas.update(import_dict)

            if isinstance(prop, RelationshipProperty):
                if prop.uselist:
                    # A one to many relationship
                    if "related_key" not in datas:
                        logger.debug("Missing infos about a relationship %s" % prop.key)
                        continue
                    datas["rel_type"] = "onetomany"
                else:
                    # A many to one relationship
                    datas.setdefault("related_key", "label")
                    # We can drop the foreignky column from the importable
                    # columns to avoid misunderstanding
                    todrop.append("%s_id" % prop.key)
                    datas["rel_type"] = "manytoone"
            else:
                column = prop.columns[0]
                if not isinstance(column.default, ColumnDefault):
                    if not column.nullable and prop.key != "id":
                        datas["mandatory"] = True

            result[datas["name"]] = datas

        for key in todrop:
            if key in result:
                ui_label = result[key].get("label")
                rel_key = key[:-3]
                if rel_key in result:
                    result[rel_key]["label"] = ui_label
                if result[key].get("mandatory"):
                    result[rel_key]["mandatory"] = True
                result.pop(key)
        return result

    def get_columns(self):
        """
        A simple getter
        """
        return self.columns

    def guess_association_dict(self, csv_datas_headers, user_defined=None):
        """
        Try to build an association dict between the header of the current csv
        file we want to import and the model we want to generate

        :param dict user_defined: An already defined association_dict we want
        to use as reference

        :returns: a dict with {headername: associated_column}
        :rtype: dict that's pickable (can be stored in the session)
        """
        result = OrderedDict()
        for header in csv_datas_headers:
            header = ascii.force_unicode(header)
            if not header:
                continue
            result[header] = None
            toguess = header.replace("*", "").lower()
            for column in list(self.columns.values()):
                name = column["name"].lower()
                label = column["label"].lower()
                if toguess in [name, label]:
                    result[header] = column["name"]

        if user_defined is not None:
            for key, value in list(user_defined.items()):
                if key in result:
                    result[key] = value

        return result

    def check_association_dict(self, association_dict=None):
        """
        Check that the provided association dict fills mandatory arguments

        :TypeError MissingMandatoryArgument: if mandatory argument is missing
            and id is in the provided keys
        """
        if association_dict is None:
            association_dict = self.association_dict
        values = list(association_dict.values())
        if "id" not in values:
            for column in list(self.columns.values()):
                if column.get("mandatory"):
                    if column["name"] not in values:
                        raise MissingMandatoryArgument(
                            MISSING_KEY_ERROR.format(column["label"], column["name"])
                        )

    def set_association_dict(self, association_dict):
        """
        Set the association dict that will be used

        :param dict association_dict: a {csv_key: model_attribute_name} dict
        """
        self.association_dict = association_dict

    def collect_args(self, csv_line, force_rel_creation=False):
        """
        Collect the arguments to be used to build the new model
        * get the value from the csv_line
        * format it thanks to the informations provided in the column info attr
        * place it in a new dict with the model attribute names as keys

        :param dict line: a csv line as a dict
        :param bool force_rel_creation: Should we try to build related
        configuration option on the fly ?
        :returns: a tuple with the args to be used for instanciation and the
            resting values
        """
        kwargs = {}
        unhandled = {}
        for csv_key, value in list(csv_line.items()):
            value = ascii.force_unicode(value)

            if csv_key == "id":
                if not value:
                    column_name = None
                else:
                    column_name = "id"
            else:
                key = ascii.force_unicode(csv_key)
                column_name = self.association_dict.get(key)
                if column_name is None:
                    column_name = self.association_dict.get(csv_key)

            if column_name is None:
                unhandled[csv_key] = value

            else:
                column = self.columns.get(column_name)
                if column is not None:
                    new_value = format_input_value(
                        value,
                        column,
                        force_rel_creation,
                    )
                    if new_value is not None:
                        kwargs[column_name] = new_value
                    else:
                        unhandled[csv_key] = value
                else:
                    kwargs[column_name] = value
        return kwargs, unhandled


def get_csv_reader(
    csv_buffer,
    delimiter=DEFAULT_DELIMITER,
    quotechar=DEFAULT_QUOTECHAR,
):
    return csv.DictReader(
        csv_buffer,
        delimiter=str(delimiter),
        quotechar=str(quotechar),
    )


class CsvImporter(object):
    """
    A csv datas importer

    * Should handle update if an id is provided
    * Should format values when generating add/update arguments
    * Should check mandatory arguments
    * Should return imported fields with errors
    * Should return non-imported datas with the appropriate id

    auto-compute relationship values
    auto-find which fields are imported

    :param class dbssession: The dbsession to use for import
    :param class factory: The type of model we want to import
    :param obj csv_buffer: A file buffer containing csv datas
    :param obj association_handler: The object handling the association
        between csv fields and model attributes
    :param str action: The action to be performed on import

            * insert: insertion of the fields is performed
            * update: we update the fields if an id key matches an existing
              object
            * override: we override the datas if an id keys matches an existing
              object
            * only_update: we update the fields if an id key matches an
            existing object and don't insert new entries
            * only_override: we override the datas if an id keys matches an
              existing object and don't insert new entries

    :param int id_key: The id key to be used (by default, we use the model's id
        to identify duplicate entries, else, we can use an external id

    :param bool force_rel_creation: Force the creation of related configuration
    elements ?

    :param dict default_values: Default arguments used to initialize new
    entries
    :param list callbacks: list of callbacks launched after importation each
    callback should take the following arguments

        action
        dbsession
        model
        updated
        args

    Usage:

        association_handler = CsvImportAssociator(UserDatas)
        importer = CsvImporter(DbSession(), UserDatas, file('users.csv', 'r'),
    """

    delimiter = ";"
    quotechar = '"'

    def __init__(
        self,
        dbsession,
        factory,
        csv_buffer,
        association_handler,
        action="only_update",
        id_key="id",
        force_rel_creation=False,
        default_values=None,
        delimiter=DEFAULT_DELIMITER,
        quotechar=DEFAULT_QUOTECHAR,
        callbacks=(),
    ):
        self.dbsession = dbsession
        self.factory = factory
        self.association_handler = association_handler
        self.in_error_lines = []
        self.unhandled_datas = []
        self.imported = []
        self.messages = []
        self.err_messages = []
        self.new_count = 0
        self.update_count = 0

        if action not in (
            "insert",
            "update",
            "override",
            "only_update",
            "only_override",
        ):
            raise KeyError(
                "The action attr should be one of " '("insert", "update", "override")'
            )

        if action == "insert":
            self.association_handler.check_association_dict()

        self.action = action
        self.id_key = id_key
        self.force_rel_creation = force_rel_creation
        self.default_init_values = default_values or {}

        self.quotechar = str(quotechar)
        self.delimiter = str(delimiter)
        self.csv_reader = get_csv_reader(csv_buffer, self.delimiter, self.quotechar)
        # Importation callbacks launched after each successfull import line
        # action
        self.callbacks = callbacks

    def import_datas(self, persist=True):
        """
        Import the datas provided in the csv_buffer as factory objects

        :param bool fake: Should datas be persisted to the database (default
        True)
        """
        for line in self.csv_reader:
            model, message = self.import_line(line, persist=persist)
            if model is None:
                self.err_messages.append(message)
            elif message is not None:
                self.messages.append(message)

        if persist:
            new_entry_msg = "{0} nouvelles entrées ont été traitées"
            update_entry_msg = "{0} entrées existantes ont été traitées"
        else:
            new_entry_msg = "{0} nouvelles entrées seront créées"
            update_entry_msg = "{0} entrées existantes seront mises à jour"

        self.messages.append(new_entry_msg.format(self.new_count))
        self.messages.append(update_entry_msg.format(self.update_count))

    def _insert(self, args, persist=True):
        """
        Insert an instance in the database

        :param dict args: The args used to instanciate our new model
        :returns: a tuple (model, updated_token) where updated_token is a
        boolean saying if it's an update
        """
        logger.debug("Inserting (1) {}".format(args))
        for key, value in list(self.default_init_values.items()):
            args[key] = value

        logger.debug("Inserting {}".format(args))
        if "id" in args:
            args.pop("id")
        model = self.factory(**args)
        if persist:
            self.dbsession.add(model)
            self.dbsession.flush()
        return model, False

    def _update(self, args, override=False, persist=True, insert=True):
        """
        Update an element in the database or insert one if no id is provided

        :param dict args: The args used to update the model
        :param bool override: should we override the existing datas ?
        :raises: sqlalchemy insert or update errors
        :returns: a tuple (model, updated_token) where updated_token is a
        boolean saying if it's an update
        """
        logger.debug("Launching update")
        logger.debug("Args :")
        logger.debug(args)
        logger.debug("Insert ? : %s" % insert)
        identification_value = args.pop(self.id_key, None)
        updated = False

        if identification_value in UNFILLED_VALUES:
            # No identification value is provided
            if not insert:
                raise MultipleInstanceFound(
                    NO_ID_KEY_ERROR.format(
                        self.id_key,
                    )
                )
            model, updated = self._insert(args, persist=persist)

        else:
            identification_column = getattr(self.factory, self.id_key)

            try:
                model = (
                    self.factory.query()
                    .filter(identification_column == identification_value)
                    .one()
                )

                for key, value in list(self.default_init_values.items()):
                    if getattr(model, key) != value:
                        log.warn("POSSIBLE BREAK IN ATTEMPT")
                        log.warn("Importation process, default values :")
                        log.warn(self.default_init_values)
                        log.warn("The model they try to edit : %s" % model.id)
                        raise Exception("POSSIBLE BREAK IN ATTEMPT !!!!!!")

                for key, value in list(args.items()):
                    if getattr(model, key) in UNFILLED_VALUES or override:
                        setattr(model, key, value)
                if persist:
                    model = self.dbsession.merge(model)
                    self.dbsession.flush()
                updated = True

            except sqlalchemy_exc.NoResultFound:
                if not insert:
                    raise MultipleInstanceFound(
                        NO_ENTRY_FOUND_ERROR.format(self.id_key, identification_value)
                    )
                # We first restore the poped identification column (if it's not
                # the id key: id key should not be set but email can be used as
                # identification key and should be set for new entries)
                if self.id_key != "id":
                    args[self.id_key] = identification_value
                model, updated = self._insert(args, persist=persist)

            except sqlalchemy_exc.MultipleResultsFound:
                raise MultipleInstanceFound(
                    MULTIPLE_ENTRY_ERROR.format(identification_value)
                )

        return model, updated

    def _only_update(self, args, persist=True):
        """
        Update an element completing its attributes
        :param dict args: The args used to update the model
        :returns: a tuple (model, updated_token) where updated_token is a
        boolean saying if it's an update
        """
        return self._update(args, override=False, persist=persist, insert=False)

    def _override(self, args, persist=True):
        """
        Update an element overriding attributes with the newly provided values
        or insert a new one if no id is provided

        :param dict args: The args used to update the model
        :returns: a tuple (model, updated_token) where updated_token is a
        boolean saying if it's an update
        """
        return self._update(args, override=True, persist=persist)

    def _only_override(self, args, persist=True):
        """
        Update an element overriding attributes with the newly provided values

        :param dict args: The args used to update the model
        :returns: a tuple (model, updated_token) where updated_token is a
        boolean saying if it's an update
        """
        return self._update(args, override=True, persist=persist, insert=False)

    def launch_callbacks(self, model, updated, args):
        """
        Launch import callbacks for the given model

        :param obj model: The current managed model
        :param bool updated: Has the model been updated
        :param dict args: The importation arguments
        """
        for callback in self.callbacks:
            callback(
                self.action,
                self.dbsession,
                model,
                updated,
                args,
            )

    def import_line(self, line, persist=True):
        """
        Import one line of a csv file
        :returns: a duple with the newly_created model (or None) and a
            message
        """
        message = None
        args, unhandled_columns = self.association_handler.collect_args(
            line,
            self.force_rel_creation,
        )

        function = getattr(self, "_{0}".format(self.action))
        logger.debug("The function we use : %s" % function)
        # Here we should handle edition
        try:
            model, updated = function(args, persist=persist)
            self.imported.append(model)
            unhandled_columns[DEFAULT_ID_LABEL] = model.id
            self.unhandled_datas.append(unhandled_columns)
            res = model
            if updated:
                self.update_count += 1
            else:
                self.new_count += 1
            if persist:
                self.launch_callbacks(model, updated, args)
        except OperationalError as error:
            import traceback

            traceback.print_exc()
            res = None
            self.in_error_lines.append(line)
            message = (
                "Erreur d'intégrité des données au lancement de la "
                "requête {} avec les paramètres {}".format(
                    error.statement, error.params
                )
            )
        except Exception as e:
            import traceback

            traceback.print_exc()
            log.exception("Erreur lors de l'import de données")
            self.in_error_lines.append(line)
            res = None
            message = str(e)
        return res, message

    def gen_csv_str(self, datas):
        """
        Generate a csv string with the given datas

        :param list datas: a list of dict representing csv rows
        :returns: a csv string
        :rtype: str
        """
        if not datas:
            result = ""
        else:
            buf = StringIO()
            fieldnames = [key for key in list(datas[0].keys()) if key != ""]
            writer = csv.DictWriter(
                buf,
                fieldnames,
                delimiter=self.delimiter,
                quotechar=self.quotechar,
                extrasaction="ignore",
                quoting=csv.QUOTE_ALL,
            )
            writer.writeheader()
            datas = ascii.to_utf8(datas)
            writer.writerows(datas)
            result = buf.getvalue()
        return result

    def log(self):
        """
        return the datas we want to log in the database (see
        models.job.CsvImportJob)
        """
        unhandled_datas_csv = self.gen_csv_str(self.unhandled_datas)
        in_error_lines_csv = self.gen_csv_str(self.in_error_lines)
        result = dict(
            unhandled_datas_csv=unhandled_datas_csv,
            in_error_lines_csv=in_error_lines_csv,
            messages=self.messages,
            error_messages=self.err_messages,
            status="done",
        )
        return result


def get_csv_importer(
    dbsession,
    model_type,
    csv_buffer,
    association_handler,
    action="insert",
    id_key="id",
    force_rel_creation=False,
    default_values=(),
    delimiter=DEFAULT_DELIMITER,
    quotechar=DEFAULT_QUOTECHAR,
):
    factory = MODELS_CONFIGURATION[model_type]["factory"]
    callbacks = MODELS_CONFIGURATION[model_type]["callbacks"]
    return CsvImporter(
        dbsession,
        factory,
        csv_buffer,
        association_handler,
        action,
        id_key,
        force_rel_creation,
        default_values,
        delimiter,
        quotechar,
        callbacks=callbacks,
    )


def get_csv_import_associator(key):
    """
    Build a csv import associator regarding the provided model
    """
    return CsvImportAssociator(
        MODELS_CONFIGURATION[key]["factory"], MODELS_CONFIGURATION[key]["excludes"]
    )


# Here we use the bind argument so that the task will be attached as a bound
# method and thus we can access attributes like request
@celery_app.task(bind=True)
def async_import_datas(
    self,
    model_type,
    job_id,
    association_dict,
    csv_filepath,
    id_key,
    action,
    force_rel_creation,
    default_values,
    delimiter,
    quotechar,
):
    """
    Launch the import of the datas provided in the csv_filepath

    :param str model_type: A handled model_type
    :param int job_id: The id of the db job object that should handle the
        return datas
    :param dict association_dict: describes the association
        csv_key<->SQLA model attribute
    :param str csv_filepath: The absolute path to the csv file
    :param str id_key: The model attribute used to handle updates
    :param str action: The name of the action we want to run
        (insert/update/override/only_update/only_override)
    :param bool force_rel_creation: Force the creation of configurable related
    elements
    :param default_values: default_values used to initialize new objects
    :param delimiter: The delimiter to use for csv parsing
    :param quotechar: The quotechar to use for csv parsing
    """
    logger.info("We are launching an asynchronous csv import")
    logger.info("  The job id : %s" % job_id)
    logger.info("  The csv_filepath : %s" % csv_filepath)
    logger.info("  The association dict : %s" % association_dict)
    logger.info("  The id key : %s" % id_key)
    logger.info("  Action : %s" % action)
    logger.info("  Default initialization values : %s" % default_values)

    utils.start_job(self.request, CsvImportJob, job_id)

    from caerp_base.models.base import DBSESSION

    try:
        associator = get_csv_import_associator(model_type)
        associator.set_association_dict(association_dict)
        csv_buffer = open(csv_filepath, "r")
        importer = get_csv_importer(
            DBSESSION(),
            model_type,
            csv_buffer,
            associator,
            action=action,
            id_key=id_key,
            force_rel_creation=force_rel_creation,
            default_values=default_values,
            delimiter=delimiter,
            quotechar=quotechar,
        )
        logger.info("Importing the datas")
        importer.import_datas()
        transaction.commit()
    except Exception as e:
        transaction.abort()
        logger.exception("The transaction has been aborted")
        utils.record_failure(CsvImportJob, job_id, [e])
    else:
        logger.info("We update the job informations")
        utils.record_completed(CsvImportJob, job_id, **importer.log())
    return ""



# -*- coding: utf-8 -*-
"""
Celery tasks related models
"""
import os
import hashlib
from datetime import (
    datetime,
    timedelta,
)
from pyramid.authorization import (
    Allow,
)
from sqlalchemy import (
    Boolean,
    Integer,
    Column,
    ForeignKey,
    String,
    DateTime,
    Text,
)
from caerp_base.models.types import (
    JsonEncodedList,
)
from caerp_base.models.mixins import (
    PersistentACLMixin,
)
from caerp_base.models.base import (
    DBBASE,
    default_table_args,
)
from sqlalchemy.orm import (
    relationship,
)


TIMEOUT = timedelta(seconds=10000)


class Job(DBBASE, PersistentACLMixin):
    """
    Base job model, used to communicate a job's status between the main pyramid
    app and celery asynchronous tasks
    """

    __tablename__ = "job"
    __table_args__ = default_table_args
    __mapper_args__ = {
        "polymorphic_on": "type_",
        "polymorphic_identity": "job",
    }
    label = "Tâche générique"
    id = Column(
        Integer,
        primary_key=True,
    )
    jobid = Column(
        String(255),
        nullable=True,
    )
    status = Column(
        String(20),
        default="planned",
    )
    created_at = Column(
        DateTime(),
        default=datetime.now,
    )
    updated_at = Column(DateTime(), default=datetime.now, onupdate=datetime.now)
    type_ = Column(
        "type_",
        String(30),
        nullable=False,
    )

    def todict(self):
        return dict(
            label=self.label,
            jobid=self.jobid,
            status=self.status,
            created_at=self.created_at.strftime("%d/%m/%Y à %H:%M"),
            updated_at=self.updated_at.strftime("%d/%m/%Y à %H:%M"),
        )

    def __json__(self, request):
        return self.todict()

    def set_owner(self, login):
        """
        Set the job owner's acls
        """
        self._acl = [
            (
                Allow,
                login,
                (
                    "add",
                    "edit",
                    "view",
                ),
            )
        ]

    def timeout(self):
        """
        Check if this element should be timeouted
        """
        result = True
        if self.status in ("planned", "running"):
            if datetime.now() - self.updated_at > TIMEOUT:
                self.status = "failed"
                if hasattr(self, "error_messages"):
                    self.error_messages = [
                        "Cette tâche a été automatiquement annulée car elle "
                        "n'a pas pu être traitée. Veulliez contacter un "
                        "administrateur en lui fournissant l'identifiant de "
                        "tâche suivant : {0}".format(self.id)
                    ]
                result = False
        return result


class CsvImportJob(Job):
    """
    Store csv importation job status
    """

    __tablename__ = "csv_import_job"
    __table_args__ = default_table_args
    __mapper_args__ = {"polymorphic_identity": "csv_import"}
    id = Column(Integer, ForeignKey("job.id"), primary_key=True)
    messages = Column(JsonEncodedList, default=None)
    error_messages = Column(JsonEncodedList, default=None)
    in_error_csv = Column(Text(), default=None)
    unhandled_datas_csv = Column(Text(), default=None)
    label = "Import de données"

    def is_not_void_str(self, value):
        """
        Return True if the string contains datas
        """
        return not (value is None or len(value) == 0)

    def todict(self):
        res = Job.todict(self)
        res["label"] = self.label
        res["messages"] = self.messages
        res["error_messages"] = self.error_messages
        res["has_errors"] = self.is_not_void_str(self.in_error_csv)
        res["has_unhandled_datas"] = self.is_not_void_str(self.unhandled_datas_csv)
        return res


class MailingJob(Job):
    """
    Store mailing job status
    """

    __tablename__ = "mailing_job"
    __table_args__ = default_table_args
    __mapper_args__ = {"polymorphic_identity": "mailing"}
    id = Column(Integer, ForeignKey("job.id"), primary_key=True)
    messages = Column(JsonEncodedList, default=None)
    error_messages = Column(JsonEncodedList, default=None)
    label = "Envoi de mail"

    def todict(self):
        res = Job.todict(self)
        res["label"] = self.label
        res["messages"] = self.messages
        res["error_messages"] = self.error_messages
        return res


class FileGenerationJob(Job):
    __tablename__ = "file_generation_job"
    __table_args__ = default_table_args
    __mapper_args__ = {"polymorphic_identity": "file_generation"}
    id = Column(Integer, ForeignKey("job.id"), primary_key=True)
    filename = Column(Text)
    error_messages = Column(JsonEncodedList, default=None)
    label = "Génération de fichier"
    # Wether to force or not the download of the result file (content-disposition: Attachment)
    force_download = Column(
        Boolean(),
        default=True,
    )

    def todict(self):
        res = Job.todict(self)
        res["filename"] = self.filename
        res["error_messages"] = self.error_messages
        res["force_download"] = self.force_download
        return res


class BulkFileGenerationJob(Job):
    """
    A single job (re)generating several caerp.models.files.File
    """

    __tablename__ = "bulk_file_generation_job"
    __table_args__ = default_table_args
    __mapper_args__ = {"polymorphic_identity": "bulk_file_generation"}
    id = Column(Integer, ForeignKey("job.id"), primary_key=True)
    error_messages = Column(JsonEncodedList, default=[])
    messages = Column(JsonEncodedList, default=None)
    results_list = Column(JsonEncodedList, default=[])
    label = "Génération de fichiers en lot"

    def todict(self):
        res = Job.todict(self)
        res["results_list"] = self.results_list
        res["error_messages"] = self.error_messages
        res["messages"] = self.messages

        return res


class MailHistory(DBBASE):
    """
    Stores the history of mail sent by our application to any company
    """

    __tablename__ = "mail_history"
    __table_args__ = default_table_args
    id = Column(Integer, primary_key=True)
    send_at = Column(
        DateTime(),
        default=datetime.now,
    )

    filepath = Column(String(255))
    md5sum = Column(String(100))
    company_id = Column(ForeignKey("company.id"), nullable=True)
    company = relationship(
        "Company",
    )

    @property
    def filename(self):
        return os.path.basename(self.filepath)


def store_sent_mail(filepath, filedatas, company_id):
    """
    Stores a sent email in the history

    :param filename: The path to the sent file
    :param filedatas: The file datas
    :param int company_id: the id of a company instance
    """
    mail_history = MailHistory(
        filepath=filepath,
        md5sum=hashlib.md5(filedatas).hexdigest(),
        company_id=company_id,
    )
    return mail_history


def check_if_mail_sent(filedatas, company_id):
    """
    Check if the given file has already been sent
    :param str filedatas: The content of a file
    :param int company_id: The id of a company
    """
    query = MailHistory.query()
    query = query.filter(MailHistory.company_id == company_id)
    md5sum = hashlib.md5(filedatas).hexdigest()
    query = query.filter(MailHistory.md5sum == md5sum)
    return query.first() is not None



# -*- coding: utf-8 -*-
import logging
from pyramid_mailer.message import (
    Attachment,
)

from caerp_base.exception import (
    MailAlreadySent,
    UndeliveredMail,
)
from caerp_base.mail import send_mail
from caerp_celery.models import (
    store_sent_mail,
    check_if_mail_sent,
)


logger = logging.getLogger(__name__)


SALARYSHEET_MAIL_MESSAGE = """Bonjour,
Vous trouverez ci-joint votre bulletin de salaire.
"""

SALARYSHEET_MAIL_SUBJECT = "Votre bulletin de salaire"


def send_salary_sheet(
    request,
    email,
    company_id,
    filename,
    filepath,
    force=False,
    message=None,
    subject=None,
):
    """
    Send a salarysheet to the given company's e-mail

    :param obj request: A pyramid request object
    :param str company_mail: The mail to send it to
    :param int company_id: The id of the associated company
    :param str filepath: The path to the filename
    :param bool force: Whether to force sending this file again
    :param str message: The mail message
    :param str subject: The mail subject
    :returns: A MailHistory instance
    :TypeError UndeliveredMail: When the company has no mail
    :TypeError MailAlreadySent: if the file has
        already been sent and no force option was passed
    """
    filebuf = open(filepath, "rb")
    filedatas = filebuf.read()

    if not force and check_if_mail_sent(filedatas, company_id):
        logger.warn("Mail already sent : mail already sent")
        raise MailAlreadySent("Mail already sent")

    filebuf.seek(0)

    if email is None:
        logger.warn(
            "Undelivered email : no mail provided for company {0}".format(company_id)
        )
        raise UndeliveredMail("no mail provided for company {0}".format(company_id))
    else:
        logger.info("Sending the file %s" % filepath)
        logger.info("Sending it to %s" % email)
        attachment = Attachment(filename, "application/pdf", filebuf)

        subject = subject or SALARYSHEET_MAIL_SUBJECT
        message = message or SALARYSHEET_MAIL_MESSAGE

        send_mail(
            request,
            email,
            message,
            subject,
            attachment,
        )
        return store_sent_mail(filepath, filedatas, company_id)


INTERNAL_ORDER_CUSTOMER_MAIL_OBJECT = """Nouvelle commande founisseur générée \
dans votre espace"""

INTERNAL_ORDER_CUSTOMER_MAIL_TMPL = """
Bonjour {customer},

L'enseigne {supplier} vous a transmis un devis 'interne'.

Une commande fournisseur contenant le devis est accessible dans votre espace
dans la section "Commande fournisseur".
"""

INTERNAL_ORDER_SUPPLIER_MAIL_OBJECT = "Votre devis a été transmis à votre \
client"
INTERNAL_ORDER_SUPPLIER_MAIL_TMPL = """
Bonjour {supplier},

Votre devis 'interne' a été transmis à l'enseigne {customer}.

Une commande fournisseur contenant le devis est accessible dans son espace dans
la section "Commande fournisseur".
"""


def send_customer_new_order_mail(request, order):
    """
    Send an email to an internal customer
    """
    customer = order.company
    supplier = order.supplier
    message = INTERNAL_ORDER_CUSTOMER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if customer.email:
        recipients = [customer.email]
        send_mail(request, recipients, message, INTERNAL_ORDER_CUSTOMER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(customer.id))


def send_supplier_new_order_mail(request, order):
    """
    Send an email to an internal supplier
    """
    customer = order.company
    supplier = order.supplier
    message = INTERNAL_ORDER_SUPPLIER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if supplier:
        recipients = [supplier.email]
        send_mail(request, recipients, message, INTERNAL_ORDER_SUPPLIER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(supplier.id))


INTERNAL_INVOICE_CUSTOMER_MAIL_OBJECT = """Nouvelle facture founisseur \
générée dans votre espace"""

INTERNAL_INVOICE_CUSTOMER_MAIL_TMPL = """
Bonjour {customer},

L'enseigne {supplier} vous a transmis une facture 'interne'.

Une facture fournisseur contenant la facture est accessible dans votre espace
dans la section "Facture fournisseur".
"""

INTERNAL_INVOICE_SUPPLIER_MAIL_OBJECT = "Votre facture a été transmis à votre \
client"
INTERNAL_INVOICE_SUPPLIER_MAIL_TMPL = """
Bonjour {supplier},

Votre facture 'interne' a été transmis à l'enseigne {customer}.

Une facture fournisseur contenant la facture est accessible dans son espace
dans la section "Facture fournisseur".
"""


def send_customer_new_invoice_mail(request, supplier_invoice):
    """
    Send an email to an internal customer
    """
    customer = supplier_invoice.company
    supplier = supplier_invoice.supplier
    message = INTERNAL_INVOICE_CUSTOMER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if customer.email:
        recipients = [customer.email]
        send_mail(request, recipients, message, INTERNAL_INVOICE_CUSTOMER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(customer.id))


def send_supplier_new_invoice_mail(request, supplier_invoice):
    """
    Send an email to an internal supplier
    """
    customer = supplier_invoice.company
    supplier = supplier_invoice.supplier
    message = INTERNAL_INVOICE_SUPPLIER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if supplier:
        recipients = [supplier.email]
        send_mail(request, recipients, message, INTERNAL_INVOICE_SUPPLIER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(supplier.id))



"""
Exemple d'écriture

.. code-block::

    N° de compte ;Libellé du compte;Date de pièce;Code JNL;Code activité;N° de pièce;Libellé entête mouvement;Libellé mouvement;Débit;Crédit;Code TVA associé;Quantité 1;Unité quantité 1 ;Prix moyen quantité 1;Unité quantité 2;Quantité 2;Prix moyen quantité 2;Sens;Solde;Code de découpe;Libellé de l'activité;Compte de contrepartie;Date 1 ;Date 2;Numéro;
    30100000;APPROVISIONNEMENTS;01/01/2023;EX;GO;;STOCK;P 23 UNITES;;688,00;;;;;;;;C;688,00;;IZA
"""

import datetime
import csv
import logging
from typing import Iterable, Optional, Generator
from pathlib import Path
from pyramid.request import Request

from caerp.compute.math_utils import str_to_float
from caerp.utils.datetimes import str_to_date

from caerp_celery.interfaces import IAccountingFileParser
from caerp_celery.parsers import BaseProducer, OperationData
from caerp_celery.parsers.sage import CsvFileParser


logger = logging.getLogger(__name__)


class IsacomptaFileParser(CsvFileParser):
    """
    Parse un fichier xlsx extrait de quadra
    """

    encoding = "iso-8859-15"

    def stream(self) -> Generator[dict, None, None]:
        with open(self.file_path, encoding=self.encoding) as fbuf:
            dialect = csv.Sniffer().sniff(fbuf.read())
            fbuf.seek(0)
            for line in csv.DictReader(fbuf, dialect=dialect):
                line = dict(
                    [(key.strip(), value.strip()) for key, value in line.items()]
                )
                yield line


class OperationProducer(BaseProducer):
    def _get_num_val(self, line: dict, key: str) -> float:
        result = 0
        val = line.get(key, "0").strip().replace(",", ".")
        result = val or 0
        if not isinstance(result, (float, int)):
            result = str_to_float(result, default=0)
        return result

    def _get_label(self, line: dict) -> str:
        label = line.get("Libellé mouvement", "").strip()
        label = label[:80]
        return label

    def _get_date(self, line: dict) -> datetime.date:
        return str_to_date(line.get("Date de pièce", "").strip()).date()

    def _stream_operation(self, line) -> Optional[OperationData]:
        analytical_account = line.get("Code activité", "").strip()
        general_account = line.get("N° de compte", "").strip()
        date = self._get_date(line)
        if not date:
            logger.error("This line has incorrect date data : %s" % line)
            return None

        label = self._get_label(line)
        credit = self._get_num_val(line, key="Crédit")
        debit = self._get_num_val(line, key="Débit")
        balance = 0
        result = OperationData(
            analytical_account=analytical_account,
            general_account=general_account,
            date=date,
            label=label,
            debit=debit,
            credit=credit,
            balance=balance,
        )
        return result

    def stream_operations(self) -> Iterable[OperationData]:
        for line in self.parser.stream():
            data = self._stream_operation(line)
            if data is not None:
                yield data


def parser_factory(context: Path, request: Request):
    return IsacomptaFileParser(context)


def producer_factory(context: IAccountingFileParser, request: Request):
    return OperationProducer(context)



import re
import hashlib
from abc import ABC, abstractmethod
import datetime
from pathlib import Path
from dataclasses import dataclass
from typing import Iterable

from caerp_celery.exception import FileNameException


@dataclass
class UploadMetaData:
    filename: str
    date: datetime.date
    filetype: str


@dataclass
class OperationData:
    analytical_account: str
    general_account: str
    date: datetime.date
    label: str
    debit: float
    credit: float
    balance: float


class BaseParser(ABC):
    filename_re = re.compile(
        "(?:general_ledger_)*(?P<year>[0-9]{4,4})_(?P<month>[0-9]{1,2})_(?P<doctype>[^.]+)",
        re.IGNORECASE,
    )

    def __init__(self, file_path: Path):
        self.file_path = file_path
        match = self.filename_re.match(self.file_path.name)
        if match is None:
            raise FileNameException(
                f"{self.file_path.name} devrait être de la forme "
                f"2022_01_grandlivre.extension ou "
                f"general_ledger_2022_01_grandlivre.extension"
            )
        else:
            self.date = datetime.date(int(match["year"]), int(match["month"]), 1)

    def metadata(self) -> UploadMetaData:
        """return Upload metadata"""
        return UploadMetaData(
            filename=self.file_path.name,
            date=self.date,
            filetype="general_ledger",
        )

    @abstractmethod
    def stream(self) -> Iterable[list]:
        """stream lines as list in the format expected by the associated producer"""
        pass


class BaseProducer(ABC):
    def __init__(self, parser: BaseParser):
        self.parser = parser

    @abstractmethod
    def stream_operations(self) -> Iterable[OperationData]:
        """
        Stream operation data used to store the accounting operations in the
        database
        """
        pass


def get_md5sum(file_path, blocksize=65536):
    """
    Return a md5 sum of the given file_path informations
    """
    hash = hashlib.md5()
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(blocksize), b""):
            hash.update(block)
    return hash.hexdigest()



import csv
import logging
from itertools import zip_longest

from typing import Iterable, List, Generator

from pathlib import Path
from pyramid.request import Request

from caerp.compute.math_utils import str_to_float
from caerp.utils.datetimes import str_to_date

from caerp_celery.interfaces import IAccountingFileParser
from caerp_celery.parsers import BaseParser, BaseProducer, OperationData


logger = logging.getLogger(__name__)


def split_list(lst, n=3):
    return zip_longest(*[iter(lst)] * n, fillvalue="")


class TxtFileParser(BaseParser):
    """
        Parse un fichier texte
    jd-3459-export-client-id-espace
        La partie contenant les écrcaerp.utils.datetimes.itures

        ECRITURES

            15/11/2021 BQ 512000 "1" "Apport capital social" D 6000,00 E  ( axe1 COOP00 )
            15/11/2021 BQ 101000 "2" "Apport capital social" C 6000,00 E  ( axe1 COOP00 )
            15/11/2021 BQ 101000 "2" "Apport capital social" C 6000,00 E  A ( axe1 COOP00 )
            ....

    """

    encoding = "iso8859-15"
    key_line = "ECRITURES"

    def _stream_operation_lines_from_file(self) -> Generator[str, None, None]:
        stream = False
        with open(self.file_path, "r", encoding=self.encoding) as fbuf:
            for line in fbuf.readlines():
                line = line.strip()
                if line.startswith("ECRITURES"):
                    stream = True
                elif stream and len(line) > 5:
                    yield line

    def stream(self) -> Generator[list, None, None]:
        for line in csv.reader(
            self._stream_operation_lines_from_file(),
            quotechar='"',
            delimiter=" ",
            skipinitialspace=True,
        ):
            yield line


class OperationProducer(BaseProducer):
    def _get_num_val(self, line: List[str], index: int) -> float:
        """
        Return the value of the given index in the given line in float format

        """
        result = 0
        if len(line) > index:
            result = line[index].strip() or 0

        return str_to_float(result, default=0.0)

    def _get_label(self, line: List[str]) -> str:
        label = line[4].strip()
        label = label[:80]
        return label

    def _get_analytical_accounts(self, line) -> List[dict]:
        """
        Collect analytical accounts from the given line

        Cas 1 :
        15/11/2021 BQ 101000 "2" "Apport capital social" C 6000,00 E  ( axe1 COOP00 )

        Cas 2 (cf le A en plus):
        15/11/2021 BQ 101000 "2" "Apport capital social" C 6000,00 E  A ( axe1 COOP00 )

        Cas 3
        15/11/2021 BQ 101000 "2" "TVA" C 6000,00 E  A ( axe1 COOP00 1000.0 axe1 COOP01 5000.0 )
        """
        # Join analytical related lines
        end_columns = " ".join(line[8:])
        if "(" in end_columns and ")" in end_columns:
            analytical_columns = [
                column
                for column in end_columns.split("(")[1].split(")")[0].split(" ")
                if column
            ]
        else:
            analytical_columns = []

        if analytical_columns.count("axe1") == 1:
            analytical_accounts = [{"account": analytical_columns[1].strip()}]
        elif analytical_columns.count("axe1") > 1:
            analytical_accounts = []
            for libelle, account, value in split_list(analytical_columns, 3):
                if account and value:
                    analytical_accounts.append(
                        {
                            "account": account.strip(),
                            "value": str_to_float(value.strip()),
                        }
                    )
        else:
            # Should not happen
            analytical_accounts = []
        return analytical_accounts

    def _stream_operation(self, line) -> Generator[OperationData, None, None]:
        print("Handling new line : %s" % line)
        if len(line) >= 12:
            analytical_accounts = self._get_analytical_accounts(line)

            general_account = line[2].strip()
            date = str_to_date(line[0].strip())
            if not date:
                logger.error("This line has incorrect datas : %s" % line)
                return None
            else:
                date = date.date()

            label = self._get_label(line)
            type_ = line[5]
            if type_ == "C":
                debit = 0
                credit = self._get_num_val(line, index=6)
            else:
                credit = 0
                debit = self._get_num_val(line, index=6)

            # Dans le cas où on a plusieurs axe sur la même ligne, on
            # splitte la ligne en plusieurs écritures analytiques
            for analytical_account_data in analytical_accounts:
                if "value" in analytical_account_data:
                    if credit == 0:
                        debit = analytical_account_data["value"]
                    else:
                        credit = analytical_account_data["value"]
                analytical_account = analytical_account_data["account"]
                balance = 0
                yield OperationData(
                    analytical_account=analytical_account,
                    general_account=general_account,
                    date=date,
                    label=label,
                    debit=debit,
                    credit=credit,
                    balance=balance,
                )

    def stream_operations(self) -> Iterable[OperationData]:
        for line in self.parser.stream():
            yield from self._stream_operation(line)


def parser_factory(context: Path, request: Request):
    return TxtFileParser(context)


def producer_factory(context: IAccountingFileParser, request: Request):
    return OperationProducer(context)



import time
import transaction
from pyramid.security import remember, forget
from sqlalchemy.orm.exc import NoResultFound
from celery.utils.log import get_task_logger


JOB_RETRIEVE_ERROR = "We can't retrieve the job {jobid}, the task will not \
be run"
TASK_RETRIEVE_ERROR = "We can't retrieve the task {}, the task will not \
be run"


def get_logger(name=""):
    return get_task_logger("celery." + name)


logger = get_logger(__name__)


# we wait max TIMEOUT seconds before considering there was an error while
# inserting job in the database on main process side
TIMEOUT = 20
# Interval of "try again"
INTERVAL = 2


def get_job(celery_request, job_model, job_id):
    """
    Return the current executed job (in caerp's sens)

    :param obj celery_request: The current celery request object
    :param obj job_model: The Job model
    :param int job_id: The id of the job

    :returns: The current job
    :raises sqlalchemy.orm.exc.NoResultFound: If the job could not be found
    """
    logger.debug("Retrieving a job with id : {0}".format(job_id))
    from caerp_base.models.base import DBSESSION

    # We try to get tje job back, waiting for the current request to be
    # finished : since we use a transaction manager, the delay call launched in
    # a view is done before the job  element is commited to the bdd (at the end
    # of the request) if we query for the job too early, the session will not
    # be able to retrieve the newly created job
    current_time = 0
    job = None
    while current_time <= TIMEOUT and job is None:
        try:
            job = DBSESSION().query(job_model).filter(job_model.id == job_id).one()
            job.jobid = celery_request.id
            if job.status not in ("planned"):
                logger.error("Job has already been launched")
                job = None
        except NoResultFound:
            transaction.abort()
            transaction.begin()
            logger.debug(" -- No job found")
            logger.exception(JOB_RETRIEVE_ERROR.format(jobid=job_id))

        if job is None:
            time.sleep(INTERVAL)
            current_time += INTERVAL

    return job


def get_task(task_id: int) -> "Task":
    """
    Retrieve a task

    :param obj celery_request: The current celery request object
    :param int task_id: The id of the Task object

    :returns: The current Task
    :raises sqlalchemy.orm.exc.NoResultFound: If the Task could not be found
    """
    logger.debug("Retrieving a task with id : {0}".format(task_id))
    from caerp_base.models.base import DBSESSION
    from caerp.models.task import (
        Task,
        Invoice,
        CancelInvoice,
    )

    # We retrieve a task back, we try several times waiting for it to be available
    # with its official_number (It should not be necessary anymore with task_tm that is
    # fired after the current web request and its underlying transaction are finished)
    current_time = 0
    task = None
    while current_time <= TIMEOUT and task is None:
        try:
            task = DBSESSION().query(Task).get(task_id)
            if task is not None and isinstance(task, (Invoice, CancelInvoice)):
                if task.official_number is None:
                    logger.error("Task has no number yet")
                    task = None
                    raise NoResultFound
        except NoResultFound:
            DBSESSION().close()
            transaction.abort()
            transaction.begin()
            logger.debug(" -- No Valid Task found")
            logger.exception(TASK_RETRIEVE_ERROR.format(task_id))

        if task is None:
            time.sleep(INTERVAL)
            current_time += INTERVAL
    logger.debug("Task {}".format(task))

    return task


def _record_running(job):
    """
    Record that a job is running
    """
    job.status = "running"
    from caerp_base.models.base import DBSESSION

    DBSESSION().merge(job)


def start_job(celery_request, job_model, job_id) -> None:
    """
    Entry point to launch when starting a job

    :param obj celery_request: The current celery request object
    :param obj job_model: The Job model
    :param int job_id: The id of the job

    :returns: The current job or None
    """
    logger.info(" Starting job %s %s" % (job_model, job_id))
    transaction.begin()
    try:
        job = get_job(celery_request, job_model, job_id)
        if job is not None:
            _record_running(job)
        else:
            raise Exception("No job found")
        transaction.commit()
    except:
        transaction.abort()
        raise Exception("Error while launching start_job")

    logger.info("Task marked as RUNNING")


def _record_job_status(job_model, job_id, status_str):
    """
    Record a status and return the job object
    """
    # We fetch the job again since we're in a new transaction
    from caerp_base.models.base import DBSESSION

    job = DBSESSION().query(job_model).filter(job_model.id == job_id).first()
    job.status = status_str
    return job


def record_failure(job_model, job_id, e=None, **kwargs):
    """
    Record a job's failure
    """
    transaction.begin()
    job = _record_job_status(job_model, job_id, "failed")
    # We append an error
    if hasattr(job, "error_messages") and e:
        job.error_messages = e

    for key, value in list(kwargs.items()):
        setattr(job, key, value)

    transaction.commit()
    logger.info("* Task marked as FAILED")


def record_completed(job_model, job_id, **kwargs):
    """
    Record job's completion and set additionnal arguments
    """
    transaction.begin()
    job = _record_job_status(job_model, job_id, "completed")
    for key, value in list(kwargs.items()):
        setattr(job, key, value)
    transaction.commit()
    logger.info("* Task marked as COMPLETED")


def check_alive():
    """
    Check the redis service is available
    """
    from pyramid_celery import celery_app
    from redis.exceptions import ConnectionError

    return_code = True
    return_msg = None
    try:
        from celery.app.control import Inspect

        insp = Inspect(app=celery_app)

        stats = insp.stats()
        if not stats:
            return_code = False
            return_msg = (
                "Le service backend ne répond pas " "(Celery service not available)."
            )
    except (Exception, ConnectionError) as e:
        return_code = False
        return_msg = "Erreur de connextion au service backend (%s)." % e

    if return_code is False:
        return_msg += " Veuillez contacter un administrateur"

    return return_code, return_msg


def set_current_user(pyramid_request, user_id):
    """
    set a user on the current request
    """
    # De telle façon qu'on ait un request.identity
    from caerp.models.user import User

    forget(pyramid_request)
    login = User.get(user_id).login.login
    remember(pyramid_request, login)



"""
Exemple de fichier xlsx

.. code-block::

    Section analytique	Général	Auxiliaire	Date	Type	Journal	Libellé	Débit	Crédit	Solde
    006000	421000	421100	31/01/2022	A	512	juvenon	2090,43	0	2090,43
"""
import logging
import datetime
from openpyxl import load_workbook
from typing import Iterable, Optional, Generator
from pathlib import Path
from pyramid.request import Request

from caerp.utils import datetimes as date_utils
from caerp.compute.math_utils import str_to_float

from caerp_celery.parsers import BaseParser, BaseProducer, OperationData
from caerp_celery.interfaces import IAccountingFileParser

logger = logging.getLogger(__name__)


class CegidFileParser(BaseParser):
    def _stream_operation_lines_from_file(self) -> Generator[dict, None, None]:
        wb = load_workbook(self.file_path)
        ws = wb.active
        rows: Generator = ws.rows
        headers = [cell.value for cell in next(rows)]
        for row in rows:
            row_data = {}
            for num_col, header in enumerate(headers):
                row_data[header] = row[num_col].value
            yield row_data

    def stream(self) -> Generator[dict, None, None]:
        for line in self._stream_operation_lines_from_file():
            yield line


class OperationProducer(BaseProducer):
    _required_keys = (
        "Section analytique",
        "Général",
        "Date",
        "Libellé",
        "Débit",
        "Crédit",
        "Solde",
    )

    def _get_date(self, line):
        date = line.get("Date", "")
        if isinstance(date, datetime.datetime):
            return date.date()
        else:
            date = date.strip()
        date_and_time = date_utils.str_to_date(date)
        if date_and_time is not None:
            return date_and_time.date()
        return None

    def _get_label(self, line) -> str:
        return line.get("Libellé", "").strip()

    def _get_num_val(self, line: dict, key: str) -> float:
        val = line.get(key, 0)
        if isinstance(val, (int, float)):
            return val
        else:
            val = val.strip()
            return str_to_float(val, default=0)

    def _stream_operation(self, line) -> Optional[OperationData]:
        for key in self._required_keys:
            if key not in line.keys():
                logger.error("This line has incorrect datas : %s" % line)
                return None
        analytical_account = line.get("Section analytique", "")
        general_account = line.get("Général", "")
        date = self._get_date(line)
        if not date:
            logger.error("This line has incorrect date data : %s" % line)
            return None
        label = self._get_label(line)
        credit = self._get_num_val(line, key="Crédit")
        debit = self._get_num_val(line, key="Débit")
        balance = 0
        result = OperationData(
            analytical_account=analytical_account,
            general_account=general_account,
            date=date,
            label=label,
            debit=debit,
            credit=credit,
            balance=balance,
        )
        return result

    def stream_operations(self) -> Iterable[OperationData]:
        for line in self.parser.stream():
            data = self._stream_operation(line)
            if data is not None:
                yield data


def parser_factory(context: Path, request: Request):
    return CegidFileParser(context)


def producer_factory(context: IAccountingFileParser, request: Request):
    return OperationProducer(context)



import transaction
from pyramid_celery import celery_app
from caerp.export.task_pdf import ensure_task_pdf_persisted
from caerp_celery.hacks import (
    setup_rendering_hacks,
)
from caerp_celery.mail import (
    send_customer_new_order_mail,
    send_supplier_new_order_mail,
    send_customer_new_invoice_mail,
    send_supplier_new_invoice_mail,
)
from caerp_celery.tasks import utils
from caerp_celery.transactional_task import task_tm
from caerp_celery.conf import get_request

logger = utils.get_logger(__name__)


@task_tm
def scheduled_render_pdf_task(document_id):
    logger.debug("Scheduling a PDF render Task for {}".format(document_id))
    document = utils.get_task(document_id)
    if document is None:
        raise Exception("Document doesn't exist in database")

    request = get_request()
    try:
        # Ensure layout_manager
        setup_rendering_hacks(request, document)
        ensure_task_pdf_persisted(document, request)
        transaction.commit()
    except Exception:
        logger.exception("Error in async_internalestimation_valid_callback")
        transaction.abort()


@task_tm
def async_internalestimation_valid_callback(document_id):
    """
    Handle the transfer of an InternalEstimation to the Client Company

    - Ensure supplier exists
    - Generates the PDF
    - Create a Supplier Order and attache the pdf file
    """
    logger.debug(
        "Async internal estimation validation callback for {}".format(document_id)
    )
    document = utils.get_task(document_id)
    if document is None:
        logger.error(
            "Document with id {} doesn't exist in database".format(document_id)
        )
        return
    else:
        logger.debug("Found the document {}".format(document))
        logger.debug(id(document))

    request = get_request()
    # Ensure layout_manager
    try:
        logger.debug("Setup rendering hacks")
        setup_rendering_hacks(request, document)
        order = document.sync_with_customer(request)
        send_customer_new_order_mail(request, order)
        send_supplier_new_order_mail(request, order)
        transaction.commit()
    except Exception:
        logger.exception("Error in async_internalestimation_valid_callback")
        transaction.abort()


@task_tm
def async_internalinvoice_valid_callback(document_id):
    """
    Handle the transfer of an InternalInvoice to the Client Company

    - Ensure supplier exists
    - Generates the PDF
    - Create a Supplier Invoice and attach the pdf file
    """
    logger.debug(
        "Async internal invoice validation callback for {}".format(document_id)
    )
    document = utils.get_task(document_id)
    if document is None:
        logger.error(
            "Document with id {} doesn't exist in database".format(document_id)
        )
        return

    pyramid_request = get_request()
    # Ensure layout_manager
    try:
        setup_rendering_hacks(pyramid_request, document)
        utils.set_current_user(pyramid_request, document.status_user_id)
        supplier_invoice = document.sync_with_customer(pyramid_request)
        send_customer_new_invoice_mail(pyramid_request, supplier_invoice)
        send_supplier_new_invoice_mail(pyramid_request, supplier_invoice)
        transaction.commit()
    except Exception:
        logger.exception("Error in async_internalinvoice_valid_callback")
        transaction.abort()
